@@ -21,9 +21,9 @@
 #     elliptic(response, model="linear", times=NULL, dose=NULL,
 #	ccov=NULL, tvcov=NULL, nest=NULL, torder=0, interaction=NULL,
 #	transform="identity", link="identity", autocorr="exponential",
-#	pell=NULL, preg=NULL, pvar=1, varfn=NULL, par=NULL, pre=NULL,
-#	delta=NULL, print.level=0, ndigit=10, gradtol=0.00001,
-#	steptol=0.00001, iterlim=100, fscale=1,
+#	pell=NULL, preg=NULL, pvar=var(y), varfn=NULL, par=NULL, pre=NULL,
+#	delta=NULL, envir=sys.frame(sys.parent()),print.level=0,
+#	ndigit=10, gradtol=0.00001, steptol=0.00001, iterlim=100, fscale=1,
 #	stepmax=10*sqrt(theta%*%theta),typsiz=abs(c(theta)))
 #
 #  DESCRIPTION
@@ -35,24 +35,24 @@
 elliptic <- function(response, model="linear", times=NULL, dose=NULL,
 	ccov=NULL, tvcov=NULL, nest=NULL, torder=0, interaction=NULL,
 	transform="identity", link="identity", autocorr="exponential",
-	pell=NULL, preg=NULL, pvar=1, varfn=NULL, par=NULL, pre=NULL,
-	delta=NULL, print.level=0, ndigit=10, gradtol=0.00001,
-	steptol=0.00001, iterlim=100, fscale=1,
+	pell=NULL, preg=NULL, pvar=var(y), varfn=NULL, par=NULL, pre=NULL,
+	delta=NULL, envir=sys.frame(sys.parent()), print.level=0,
+	ndigit=10, gradtol=0.00001, steptol=0.00001, iterlim=100, fscale=1,
 	stepmax=10*sqrt(theta%*%theta),typsiz=abs(c(theta))){
 plra <- function(theta){
-	if(mdl==2)mu <- model(theta)
-	if(cvar==1)var <- varfn(theta[n1:n2])
+	if(mdl==2)mu <- mu1(theta)
+	if(cvar==1)varn <- sh1(theta[n1:n2])
 	z <- .Fortran("plra",
 		theta=as.double(theta),
 		like=double(1),
 		as.double(rxl),
 		x=as.double(times),
 		as.double(y),
-		tvcov=as.double(zna$tvcov$tvcov),
-		ccov=as.double(zna$ccov$ccov),
+		tvcov=as.double(response$tvcov$tvcov),
+		ccov=as.double(response$ccov$ccov),
 		dose=as.double(dose),
-		nobs=as.integer(zna$response$nobs),
-		nest=as.integer(zna$response$nest),
+		nobs=as.integer(response$response$nobs),
+		nest=as.integer(response$response$nest),
 		lnest=as.integer(lnest),
 		dev=double(nm),
 		nind=as.integer(nind),
@@ -73,29 +73,30 @@ plra <- function(theta){
 		model=as.integer(mdl),
 		ar=as.integer(ar),
 		tvc=as.integer(tvc),
-		beta=as.double(beta),
-		betacov=betacov,
+		beta=double(npv2),
+		betacov=double(npv2*npv2),
 		v=double(nld*nld),
 		sigsq=double(nld),
 		ey=double(nld),
 		tb=double(npvl),
 		as.double(mu),
-		as.double(var))
+		as.double(varn),
+		DUP=F)
 	list(like=z$like,res=z$dev,beta=z$beta,betacov=z$betacov)}
 plral <- function(theta){
-	if(mdl==2)mu <- model(theta)
-	if(cvar==1)var <- varfn(theta[n1:n2])
+	if(mdl==2)mu <- mu1(theta)
+	if(cvar==1)varn <- sh1(theta[n1:n2])
 	z <- .Fortran("plra",
 		theta=as.double(theta),
 		like=double(1),
 		as.double(rxl),
 		x=as.double(times),
 		as.double(y),
-		tvcov=as.double(zna$tvcov$tvcov),
-		ccov=as.double(zna$ccov$ccov),
+		tvcov=as.double(response$tvcov$tvcov),
+		ccov=as.double(response$ccov$ccov),
 		dose=as.double(dose),
-		nobs=as.integer(zna$response$nobs),
-		nest=as.integer(zna$response$nest),
+		nobs=as.integer(response$response$nobs),
+		nest=as.integer(response$response$nest),
 		lnest=as.integer(lnest),
 		dev=double(nm),
 		nind=as.integer(nind),
@@ -116,98 +117,217 @@ plral <- function(theta){
 		model=as.integer(mdl),
 		ar=as.integer(ar),
 		tvc=as.integer(tvc),
-		beta=as.double(beta),
-		betacov=betacov,
+		beta=double(npv2),
+		betacov=double(npv2*npv2),
 		v=double(nld*nld),
 		sigsq=double(nld),
 		ey=double(nld),
 		tb=double(npvl),
 		as.double(mu),
-		as.double(var))
+		as.double(varn),
+		DUP=F)
 	z$like}
 call <- sys.call()
-if(!is.function(model)&&!missing(model))
+if(!is.function(model)&&!inherits(model,"formula")&&!is.null(model))
 	model <- match.arg(model,c("linear","logistic","pkpd"))
 tmp <- c("exponential","gaussian","cauchy","spherical","IOU")
 ar <- match(autocorr <- match.arg(autocorr,tmp),tmp)
 transform <- match.arg(transform,c("identity","exp","square","sqrt","log"))
 tmp <- c("identity","exp","square","sqrt","log")
 lnk <- match(link <- match.arg(link,tmp),tmp)
+npar <- ifelse(!is.null(par),1,0)
+tvc <- ifelse(!is.null(tvcov),1,0)
+if(!inherits(response,"repeated")){
+	if(!inherits(response,"response"))response <- restovec(response,times,nest=nest,delta=delta)
+	if(is.null(ccov))nccov <- 0
+	else {
+		if(!inherits(ccov,"tccov")){
+			ccname <- paste(deparse(substitute(ccov)))
+			if((is.matrix(ccov)&&is.null(colnames(ccov)))){
+				ccname <- paste(deparse(substitute(ccov)))
+				if(ncol(ccov)>1){
+					tmp <- NULL
+					for(i in 1:ncol(ccov))tmp <- c(tmp,paste(ccname,i,sep=""))
+					ccname <- tmp}}
+			ccov <- tcctomat(ccov,names=ccname)}
+		nccov <- ncol(ccov$ccov)}
+	if(is.null(tvcov))tvc <- 0
+	else {
+		if(!inherits(tvcov,"tvcov")){
+			tvcname <- paste(deparse(substitute(tvcov)))
+			if(is.list(tvcov)&&ncol(tvcov[[1]])>1){
+				if(is.null(colnames(tvcov[[1]]))){
+					tvcname <- paste(deparse(substitute(tvcov)))
+					tmp <- NULL
+					for(i in 1:ncol(tvcov[[1]]))tmp <- c(tmp,paste(tvcname,i,sep=""))
+					tvcname <- tmp}
+				else tvcname <- colnames(tvcov[[1]])}
+			tvcov <- tvctomat(tvcov,names=tvcname)}
+		tvc <- ncol(tvcov$tvcov)}
+	response <- rmna(response=response, tvcov=tvcov, ccov=ccov)
+	if(!is.null(ccov))rm(ccov)
+	if(!is.null(tvcov))rm(tvcov)}
+else {
+	nccov <- if(is.null(response$ccov$ccov)) 0
+		 else  ncol(response$ccov$ccov)
+	tvc <- if(is.null(response$tvcov$tvcov)) 0
+		 else  ncol(response$tvcov$tvcov)}
+if((inherits(envir,"repeated")&&
+	(length(response$response$nobs)!=length(envir$response$nobs)||
+	any(response$response$nobs!=envir$response$nobs)))||
+	(inherits(envir,"tvcov")&&
+	(length(response$response$nobs)!=length(envir$tvcov$nobs)||
+	any(response$response$nobs!=envir$tvcov$nobs))))
+	stop("response and envir objects are incompatible")
+full <- !is.null(response$ccov$ccov)
+y <- response$response$y
+times <- response$response$times
+nind <- length(response$response$nobs)
+nld <- max(response$response$nobs)
 mu <- NULL
-var <- NULL
-n1 <- length(preg)+1
-n2 <- length(c(preg,pvar))
-npell <- !missing(pell)
+varn <- NULL
+npell <- !is.null(pell)
 if(npell&&pell<=0)stop("The elliptic power parameter must be positive.")
-npvar <- length(pvar)
-if(!missing(pre))npre <- length(pre)
+if(!is.null(pre))npre <- length(pre)
 else npre <- 0
-npar <- ifelse(!missing(par),1,0)
-tvc <- ifelse(!missing(tvcov),1,0)
-if(is.function(model)){
-	if(is.null(preg))stop("when model is a function, preg must be supplied")
-	mdl <- 2}
+mu3 <- sh3 <- name <- NULL
+if(inherits(envir,"repeated")||inherits(envir,"tccov")){
+	type <- if(inherits(envir,"repeated"))"repeated"
+		else if(inherits(envir,"tccov"))"tccov"
+		else "tvcov"
+	name <- paste(deparse(substitute(envir)))
+	if(inherits(mu,"formula")){
+		mu3 <- finterp(mu)
+		class(mu) <- c(class(mu),type)}
+	else if(is.function(model)){
+		mu3 <- model
+		attributes(mu3) <- attributes(fnenvir(model))
+		class(model) <- type
+		model <- fnenvir(model,envir=envir,name=name)}
+	if(inherits(shape,"formula")){
+		sh3 <- finterp(shape)
+		class(shape) <- c(class(shape),type)}
+	else if(is.function(varfn)){
+		sh3 <- varfn
+		attributes(sh3) <- attributes(fnenvir(varfn))
+		class(varfn) <- type
+		varfn <- fnenvir(varfn,envir=envir,name=name)}}
+npr <- length(preg)
+if(inherits(model,"formula")){
+	mu2 <- finterp(model,envir=envir,name=name)
+	npt1 <- length(attr(mu2,"parameters"))
+	if(is.matrix(attr(mu2,"model"))){
+		if(all(dim(attr(mu2,"model"))==1)){
+			mu1 <- function(p) p[1]*rep(1,n)
+			attributes(mu1) <- attributes(mu2)
+			mu2 <- NULL}}
+	else {
+		if(npr!=npt1){
+			cat("\nParameters are ")
+			cat(attr(mu2,"parameters"),"\n")
+			stop(paste("preg should have",npt1,"estimates"))}
+		if(is.list(preg)){
+			if(!is.null(names(preg))){
+				o <- match(attr(mu2,"parameters"),names(preg))
+				preg <- unlist(preg)[o]
+				if(sum(!is.na(o))!=length(preg))stop("invalid estimates for mu - probably wrong names")}
+			else preg <- unlist(preg)}}
+	if(!is.null(mu2)){
+		if(inherits(envir,"tccov")){
+			cv <- covind(response)
+			mu1 <- function(p) mu2(p)[cv]
+			attributes(mu1) <- attributes(mu2)}
+		else {
+			mu1 <- mu2
+			rm(mu2)}}}
+else if(is.function(model))mu1 <- model
+else mu1 <- NULL
+if(!is.null(mu1)&&is.null(attributes(mu1))){
+	attributes(mu1) <- if(is.function(model)){
+		if(!inherits(model,"formulafn"))attributes(fnenvir(model))
+		else attributes(model)}
+		else attributes(fnenvir(mu1))}
+nlp <- if(is.function(mu1))length(attr(mu1,"parameters"))
+	else if(is.null(mu1))NULL
+	else npt1
+if(!is.null(nlp)&&nlp!=npr)
+	stop(paste("preg should have",nlp,"initial estimates"))
+if(is.function(mu1))mdl <- 2
 else if(model=="linear"){
 	mdl <- 1
 	torder <- torder+1}
 else if(model=="logistic")mdl <- 3
 else if(model=="pkpd")mdl <- 4
-linear <- NULL
-if(missing(ccov)){
-	nccov <- 0
-	zc <- NULL}
-else {
-	if(inherits(ccov,"tccov"))zc <- ccov
+npvar <- length(pvar)
+if(inherits(varfn,"formula")){
+	sh2 <- finterp(varfn,envir=envir,name=name)
+	npt2 <- length(attr(sh2,"parameters"))
+	if(is.matrix(attr(sh2,"model"))){
+		if(all(dim(attr(sh2,"model"))==1)){
+			sh1 <- function(p) p[1]*rep(1,n)
+			attributes(sh1) <- attributes(sh2)
+			sh2 <- NULL}}
 	else {
-		if(is.vector(ccov,mode="double")||(is.matrix(ccov)&&is.null(colnames(ccov))))
-			ccname <- paste(deparse(substitute(ccov)))
-		else ccname <- NULL
-		zc <- tcctomat(ccov,names=ccname)}
-	nccov <- ncol(zc$ccov)}
-if(mdl==1&&!missing(interaction)){
+		if(npvar!=npt2){
+			cat("\nParameters are ")
+			cat(attr(sh2,"parameters"),"\n")
+			stop(paste("pvar should have",npt2,"estimates"))}
+		if(is.list(pvar)){
+			if(!is.null(names(pvar))){
+				o <- match(attr(sh2,"parameters"),names(pvar))
+				pvar <- unlist(pvar)[o]
+				if(sum(!is.na(o))!=length(pvar))stop("invalid estimates for varfn - probably wrong names")}
+			else pvar <- unlist(pvar)}}
+	if(!is.null(sh2)){
+		if(inherits(envir,"tccov")){
+			cv <- covind(response)
+			sh1 <- function(p) sh2(p)[cv]
+			attributes(sh1) <- attributes(sh2)}
+		else {
+			sh1 <- sh2
+			rm(sh2)}}}
+else if(is.function(varfn))sh1 <- varfn
+else sh1 <- NULL
+if(!is.null(sh1)&&is.null(attributes(sh1)))
+	attributes(sh1) <- if(is.function(varfn)){
+		if(!inherits(varfn,"formulafn"))attributes(fnenvir(varfn))
+		else attributes(varfn)}
+		else attributes(fnenvir(sh1))
+nlp <- if(is.function(varfn))length(attr(sh1,"parameters"))
+	else if(is.null(varfn))NULL
+	else npt2
+if(!is.null(nlp)&&nlp!=npvar)
+	stop(paste("pvar should have",nlp,"initial estimates"))
+if(mdl==1&&!is.null(interaction)){
 	if(length(interaction)!=nccov)
 		stop(paste(nccov,"interactions with time must be specified"))
 	else if(any(interaction>torder-1))
 		stop(paste("Interactions can be at most of order ",torder-1))
 	else interaction <- interaction+1}
 else interaction <- rep(1,nccov)
-if(mdl==4&&tvc==0&&missing(dose))stop("Doses required for PKPD model")
-full <- !is.null(ccov)
-if(inherits(response,"response"))zr <- response
-else zr <- restovec(response,times,nest=nest,delta=delta)
-if(!missing(tvcov)){
-	if(inherits(tvcov,"tvcov"))zt <- tvcov
-	else {
-		if(is.matrix(tvcov)||is.data.frame(tvcov)||
-			(is.list(tvcov)&&is.null(colnames(tvcov[[1]]))))
-			tvcname <- paste(deparse(substitute(tvcov)))
-		else tvcname <- NULL
-		zt <- tvctomat(tvcov,names=tvcname)}
-	tvc <- ncol(zt$tvcov)}
-else {
-	tvc <- 0
-	zt <- NULL}
-zna <- rmna(response=zr, tvcov=zt, ccov=zc)
-rm(zr,zt,zc)
-y <- zna$response$y
-times <- zna$response$times
-nind <- length(zna$response$nobs)
-nld <- max(zna$response$nobs)
-if(!is.null(zna$response$nest))lnest <- max(zna$response$nest)
+if(mdl==4&&tvc==0&&is.null(dose))stop("Doses required for PKPD model")
+n1 <- length(preg)+1
+n2 <- length(c(preg,pvar))
+if(!is.null(response$response$nest))lnest <- max(response$response$nest)
 else {
 	lnest <- 0
-	zna$response$nest <- rep(1,length(y))}
-if(mdl==2&&length(model(preg))!=sum(zna$response$nobs))
+	response$response$nest <- rep(1,length(y))}
+if(mdl==2&&length(mu1(preg))!=sum(response$response$nobs))
 	stop("The mean function must provide an estimate for each observation")
-if(mdl==1){
-	ave <- mean(times)
-	times <- times-ave}
-else ave <- 0
+if(!is.null(times)){
+	if(mdl==1){
+		ave <- mean(times)
+		times <- times-ave}
+	else ave <- 0}
+else {
+	if(!is.null(par))stop("No times. AR cannot be fitted")
+	else if(torder>0)stop("No times. Time trends cannot be fitted.")
+	ave <- 0}
 if(full){
 	rxl <- NULL
 	for(i in 1:nind){
 		tmp <- 1
-		if(full)for(j in 1:ncol(zna$ccov$ccov))tmp <- tmp+zna$ccov$ccov[i,j]*2^(j-1)
+		if(full)for(j in 1:ncol(response$ccov$ccov))tmp <- tmp+response$ccov$ccov[i,j]*2^(j-1)
 		rxl <- c(rxl,tmp)}
 	nxrl <- max(rxl)
 	p <- preg}
@@ -215,7 +335,7 @@ else {
 	nxrl <- 1
 	rxl <- matrix(1,ncol=1,nrow=nind)
 	p <- NULL}
-nm <- sum(zna$response$nobs)
+nm <- sum(response$response$nobs)
 if(transform=="identity")jacob <- 0
 else if(transform=="exp"){
 	jacob <- -sum(y)
@@ -232,39 +352,35 @@ else {
 	else if(transform=="log"){
 		jacob <- sum(log(y))
 		y <- log(y)}}
-if(!is.null(zna$response$delta)){
-	if(length(zna$response$delta)==1)jacob <- jacob-nm*log(zna$response$delta)
-	else jacob <- jacob -sum(log(zna$response$delta))}
+if(!is.null(response$response$delta)){
+	if(length(response$response$delta)==1)jacob <- jacob-nm*log(response$response$delta)
+	else jacob <- jacob -sum(log(response$response$delta))}
 if(mdl==1){
-	if(lnk==1&&missing(varfn)){
+	if(lnk==1&&is.null(varfn)){
 		npvl <- torder+sum(interaction)+tvc
-		npv <- 0
-		beta <- double(npvl)
-		betacov <- matrix(0,ncol=npvl,nrow=npvl)}
+		npv <- 0}
 	else {
 		npv <- torder+sum(interaction)+tvc
 		npvl <- 0
-		betacov <- NULL
-		if(missing(preg)){
-			if(lnk!=1)stop(paste("Initial regression estimates must be supplied for the linear model with ",link,"link"))
-			else stop("Initial regression estimates must be supplied for the linear model when there is a variance function")}
-		else beta <- preg}}
+		if(is.null(preg)){
+			cat("Initial regression estimates must be supplied ")
+			if(lnk!=1)stop(paste("for the linear model with ",link,"link"))
+			else stop("for the linear model when there is a variance function")}}}
 else {
 	npvl <- 0
-	beta <- 0
-	betacov <- NULL
 	if(mdl==2)npv <- length(preg)
-	else if(mdl==3)npv <- ifelse(tvc==0,4*nxrl,4+nxrl-1)
+	else if(mdl==3)npv <- if(tvc==0)4*nxrl else 4+nxrl-1
 	else if(mdl==4){
 		if(!is.function(varfn))npv <- 2+nxrl
 		else npv <- 3}}
+npv2 <- max(npv,npvl)
 if(length(preg)!=npv)
-	stop(paste(npv,"initial parameter estimates for growth curve must be supplied"))
-if(missing(varfn))cvar <- 0
-else if(is.function(varfn)){
+	stop(paste(npv,"initial parameter estimates for the mean regression must be supplied"))
+if(is.null(varfn))cvar <- 0
+else if(is.function(sh1)){
 	cvar <- 1
-	if(length(varfn(pvar))!=sum(zna$response$nobs))
-		stop("The variance function must provide an estimate for each observation")}
+	if(length(sh1(pvar))!=sum(response$response$nobs))
+		stop("The variance function or formula must provide an estimate for each observation")}
 else if(varfn=="identity"){
 	if(any(y<0))warning("identity variance function not recommended with negative responses")
 	cvar <- 2}
@@ -306,13 +422,14 @@ dimnames(nlcorr) <- list(seq(1,np-npvl),seq(1,np-npvl))
 betase <- betacorr <- NULL
 if(mdl==1){
 	if(lnk==1&&cvar==0){
+		betacov <- matrix(z$betacov,ncol=npvl)
 		bt <- exp(p[length(p)])
-		if(npell>0)z$betacov <- z$betacov/(npvl*gamma(npvl/(2*bt)))*
+		if(npell>0)betacov <- betacov/(npvl*gamma(npvl/(2*bt)))*
 			2^(1/bt)*gamma((npvl+2)/(2*bt))
-		if(npvl==1)betase <- sqrt(z$betacov)
-		else if(npvl>1)betase <- sqrt(diag(z$betacov))
+		if(npvl==1)betase <- sqrt(betacov)
+		else if(npvl>1)betase <- sqrt(diag(betacov))
 		if(npvl>1){
-			betacorr <- z$betacov/(betase%o%betase)
+			betacorr <- betacov/(betase%o%betase)
 			dimnames(betacorr) <- list(seq(1,npvl),seq(1,npvl))}}
 	else {
 		betase <- nlse[1:npv]
@@ -320,6 +437,7 @@ if(mdl==1){
 		if(npvl>1){
 			betacorr <- nlcorr[1:npv,1:npv]
 			dimnames(betacorr) <- list(seq(1,npv),seq(1,npv))}}}
+else betacov <- NULL
 sigsq <- p[(npv+1):(npv+npvar)]
 if(!(mdl==4&npvar==4))sigsq <- exp(sigsq)
 if(npre>0)tausq <- exp(p[(npv+npvar+1):(npv+npvar+npre)])
@@ -328,16 +446,21 @@ if(npar>0){
 	rho <- exp(p[npv+npvar+npre+1])
 	if(ar!=5)rho <- rho/(1+rho)}
 else rho <- 0
+if(!is.null(mu3))mu1 <- mu3
+if(!is.null(sh3))sh1 <- sh3
 z <- list(
 	call=call,
 	model=model,
+	mu1=mu1,
+	varfn=varfn,
+	sh1=sh1,
 	autocorr=autocorr,
-	response=zna$response,
+	response=response$response,
 	transform=transform,
 	torder=torder-1,
 	interaction=interaction-1,
-	ccov=zna$ccov,
-	tvcov=zna$tvcov,
+	ccov=response$ccov,
+	tvcov=response$tvcov,
 	full=full,
 	link=link,
 	maxlike=like,
@@ -348,12 +471,11 @@ z <- list(
 	npv=npv,
 	npvl=npvl,
 	npvar=npvar,
-	varfn=varfn,
 	npar=npar,
 	npre=npre,
 	coefficients=p,
 	beta=z$beta,
-	betacov=z$betacov,
+	betacov=betacov,
 	betacorr=betacorr,
 	betase=betase,
 	nlse=nlse,
@@ -377,12 +499,13 @@ print.elliptic <- function(z, digits = max(3, .Options$digits - 3)) {
 	if(!is.null(z$ccov$ccov))nccov <- ncol(z$ccov$ccov)
 	else nccov <- 0
 	if(z$npell==0)cat("\nMultivariate normal distribution\n")
-	else cat("\nMultivariate elliptically contoured distribution\n")
-	cat("\nCall:\n",deparse(z$call),"\n\n",sep="")
+	else cat("\nMultivariate elliptically-contoured distribution\n")
+	cat("\nCall:",deparse(z$call),sep="\n")
+	cat("\n")
 	if(z$code>2)cat("Warning: no convergence - error",z$code,"\n\n")
 	cat("Number of subjects    ",length(z$response$nobs),"\n")
 	cat("Number of observations",length(z$response$y),"\n")
-	if(!is.function(z$model)){
+	if(!inherits(z$mu1,"formulafn")){
 		if(z$model=="linear"){
 			if(z$torder>0){
 				cat("\nPolynomial model\n")
@@ -397,15 +520,22 @@ print.elliptic <- function(z, digits = max(3, .Options$digits - 3)) {
 	cat("AIC               ",z$aic,"\n")
 	cat("Iterations        ",z$iterations,"\n\n")
 	tvc <- !is.null(z$tvcov)
-	if(!is.function(z$model))cat("Location parameters\n")
-	if(is.language(z$ccov$linear))
-		cat("Formula: ",deparse(z$ccov$linear),"\n")
-	if(is.function(z$model)){
-		t <- deparse(z$model)
+	if(!inherits(z$mu1,"formulafn"))
+		cat("Location parameters\n")
+	if(inherits(z$ccov$linear,"formula"))
+		cat("Linear part: ",deparse(z$ccov$linear),sep="\n")
+	if(inherits(z$mu1,"formulafn")){
 		cat("Parameters for location function\n")
-		cat(t[2:length(t)],"",sep="\n")
+		if(!is.null(attr(z$mu1,"formula")))cat(deparse(attr(z$mu1,"formula")),sep="\n")
+		else if(!is.null(attr(z$mu1,"model"))){
+			t <- deparse(attr(z$mu1,"model"))
+			t[1] <- sub("expression\\(","",t[1])
+			t[length(t)] <- sub("\\)$","",t[length(t)])
+			cat(t,sep="\n")}
+		cname <- if(is.matrix(attr(z$mu1,"model")))
+				colnames(attr(z$mu1,"model"))
+			else attr(z$mu1,"parameters")
 		coef.table <- cbind(z$coef[1:z$npv], z$nlse[1:z$npv])
-		cname <- paste("p",1:z$npv,sep="")
 		dimnames(coef.table) <- list(cname, c("estimate", "se"))
 		print.default(coef.table, digits=digits, print.gap=2)}
 	else if(z$model=="linear"){
@@ -455,15 +585,21 @@ print.elliptic <- function(z, digits = max(3, .Options$digits - 3)) {
 				coef.table <- cbind(z$coef[i+3],z$nlse[i+3])
 				dimnames(coef.table) <- list(cname[3], c("estimate", "se"))
 				print.default(coef.table, digits=digits, print.gap=2)}}}
-	if(is.function(z$varfn)){
-		cat("\nVariance parameters for variance function\n")
-		t <- deparse(z$varfn)
-		cat(t[2:length(t)],"",sep="\n")
+	if(inherits(z$sh1,"formulafn")){
+		cat("\nParameters for variance function\n")
+		if(!is.null(attr(z$sh1,"formula")))cat(deparse(attr(z$sh1,"formula")),sep="\n")
+		else if(!is.null(attr(z$sh1,"model"))){
+			t <- deparse(attr(z$sh1,"model"))
+			t[1] <- sub("expression\\(","",t[1])
+			t[length(t)] <- sub("\\)$","",t[length(t)])
+			cat(t,sep="\n")}
+		cname <- if(is.matrix(attr(z$sh1,"model")))
+				colnames(attr(z$sh1,"model"))
+			else attr(z$sh1,"parameters")
 		coef.table <- cbind(z$coef[(z$npv+1):(z$npv+z$npvar)], z$nlse[(z$npv+1):(z$npv+z$npvar)])
-		cname <- paste("p",1:z$npvar,sep="")
 		dimnames(coef.table) <- list(cname, c("estimate", "se"))
 		print.default(coef.table, digits=digits, print.gap=2)}
-	else if(!is.function(z$model)&&z$model=="pkpd"&&z$npvar==4){
+	else if(!is.function(z$model)&&!inherits(z$mu1,"formulafn")&&z$model=="pkpd"&&z$npvar==4){
 		cat("\nVariance parameters\n")
 		coef.table <- cbind(z$sigsq, z$nlse[(z$npv+1):(z$npv+4)])
 		cname <- c("log k_a","log k_e","log V","power")
@@ -498,7 +634,7 @@ print.elliptic <- function(z, digits = max(3, .Options$digits - 3)) {
 	if(z$npell==1){
 		cat("\nElliptical distribution power parameter\n")
 		coef.table <- cbind(z$coef[z$np-z$npvl],z$nlse[z$np-z$npvl],exp(z$coef[z$np-z$npvl]))
-		dimnames(coef.table) <- list(1,c("estimate","se","power"))
+		dimnames(coef.table) <- list("",c("estimate","se","power"))
 		print.default(coef.table, digits=digits, print.gap=2)}
 	if(z$np-z$npvl>1){
 		cat("\nCorrelation matrix of nonlinear parameters\n")
